/******************************************************************************/
/* gillespie algorithm                                                        */
/* (c) 2007 jamie twycross, jpt AT cs.nott.ac.uk                              */
/* released under GNU GPL v2                                                  */
/******************************************************************************/

/******************************************************************************/
/* to compile:                                                                */
/*      # make clean all                                                      */
/* needs libgsl (GNU Scientific Library - GSL) development files installed    */
/* libgsl is used for efficient + reliable random number generation           */
/*                                                                            */
/* to run:                                                                    */
/*      # ./gillespie OUTFILE TIMESPAN [SEED]                                 */
/* where:                                                                     */
/*      OUTFILE is a file to save the data produced by the simulation         */
/*      TIMESPAN is the length of time (seconds) to run the simulation for    */
/*      SEED is an optional seed for the random number generator - otherwise  */
/*           a random seed is used                                            */
/* for example:                                                               */
/*      # ./gillespie lotka.data 20                                           */
/*                                                                            */
/* the output file is composed of space seperated columns:                    */
/*      ITERATION_NUMBER TIME TAU MU SPECIES_1 SPECIES_2 ...                  */
/* where:                                                                     */
/*      TAU and MU are the time to and number of the next reaction            */
/*      SPECIES_1, SPECIES_2, ... are the population levels for each species  */
/*                                                                            */
/* to quickly view the resulting output, run gnuplot:                         */
/*      # gnuplot                                                             */
/* then (for a phase-space plot):                                             */
/*      gnuplot> plot "lotka.data" using 5:6                                  */
/* or (for time-series plots of the levels of species 1 and 2):               */
/*      gnuplot> plot "lotka.data" using 2:5                                  */
/*      gnuplot> plot "lotka.data" using 2:6                                  */
/*                                                                            */
/* three reactions have been implemented from Gillespie's 1977 paper:         */
/*      Daniel T. Gillespie. Exact Stochastic Simulation of Coupled Chemical  */
/*      Reactions. Journal of Physical Chemistry, 81(25):2340-2361, 1977.     */
/* see the "user-defined functions" section and main() below                  */
/******************************************************************************/

#include <stdlib.h>
#include <err.h>
#include <gsl/gsl_rng.h>
#include <math.h>
#include <time.h>

#define MAX_SPECIES	100  /* max number of chemical species */

/* a chemical system */
typedef struct Chemical_System {
	/* user-defined parameters */
	unsigned int M;  /* number of chemical reactions */
	double *c;  /* stocastic reaction constants */
	unsigned int N;  /* number of chemical species */
	unsigned long int *X;  /* population level for each species */
	gsl_rng *rng;  /* random number generator */
	unsigned int (*h)(struct Chemical_System *, \
			const unsigned int);  /* reactant combination function */
	void (*R)(struct Chemical_System *, \
			const unsigned int);  /* reaction function */
	
	double t;  /* time */
	unsigned long int n;  /* number of iterations */
	double tau;  /* value of tau calculated */
	unsigned int mu;  /* reaction which occured */
	double *a;  /* used for intermediate calculations */
} Chemical_System;

/*********************/
/* generic functions */
/*********************/

/* initialise memory for chemical system */
Chemical_System *initialise_chemical_system(const unsigned int nr, \
		const unsigned int ns, const unsigned long seed)
{
	Chemical_System *cs;
	
	if(!(cs = malloc(sizeof(Chemical_System))))
		errx(EXIT_FAILURE, "couldn't allocate memory for chemical system");
	cs->t = 0.0;  /* set time to zero */
	cs->n = 0;  /* set number of iterations */
	cs->M = nr;  /* set number of reactions */
	cs->N = ns;  /* set number of species */
	/* allocate memory for stocastic reaction constants */
	if(!(cs->c = calloc(cs->M, sizeof(double))))
		errx(EXIT_FAILURE, "couldn't allocate memory for reaction constants");
	/* allocate memory for species population levels */
	if(!(cs->X = calloc(cs->N, sizeof(unsigned long int))))
		errx(EXIT_FAILURE, "couldn't allocate memory for population levels");
	/* allocate memory for intermediate calculations */
	if(!(cs->a = calloc(cs->M, sizeof(double))))
		errx(EXIT_FAILURE, "couldn't allocate memory for intermediate array");
	/* initialise random number generator */
	cs->rng = gsl_rng_alloc(gsl_rng_mt19937);
	gsl_rng_set(cs->rng, seed);

	return cs;
}

/* perform a single iteration of the gillespie algorithm */
void gillespie_algorithm(Chemical_System *cs)
{
	register unsigned int i, mu;
	static double a0, r1, r2, s1;
	
	/* step 1 */
	a0 = 0.0;
	for(i = 0; i < cs->M; i++) {
		cs->a[i] = cs->h(cs, i + 1) * cs->c[i];
		a0 += cs->a[i];
	}
	if(a0 == 0.0) {  /* stopping criteria */
		cs->t += cs->tau;
		cs->mu = 0;
		cs->n++;
		return;
	}
	
	/* step 2 */
	r1 = gsl_rng_uniform(cs->rng);
	r2 = gsl_rng_uniform(cs->rng);
	cs->tau = -1.0 * log(r1) / a0;
	s1 = r2 * a0;
	for(mu = 0; mu < cs->M; mu++) {
		s1 -= cs->a[mu]; 
		if(s1 <= 0.0)
			break;
	}
	cs->mu = mu + 1;
	
	/* step 3 */
	cs->t += cs->tau;
	cs->R(cs, cs->mu);
	cs->n++;
}

/* output information on current state of chemical system */
void output_chemical_system(Chemical_System *cs, FILE *fp)
{
	unsigned int i;
	fprintf(fp, "%ld %f %f %d", cs->n, cs->t, cs->tau, cs->mu);
	for(i = 0; i < cs->N; i++)
		fprintf(fp, " %ld", cs->X[i]);
	fprintf(fp, "\n");
}

/* check if any chemicals left in system */
int empty_chemical_system(Chemical_System *cs)
{
	register unsigned int i;
	
	for(i = 0; i < cs->N; i++)
		if(cs->X[i] > 0)
			return 0;
	return 1;
}

/**************************/
/* user-defined functions */
/**************************/

/* irreversible isomerization (reaction 22) */

unsigned int h_system_1(Chemical_System *cs, const unsigned int i)
{
	return cs->X[0];
}

void R_system_1(Chemical_System *cs, const unsigned int i)
{
	if(cs->X[0] > 0)
		cs->X[0]--;
}

Chemical_System *initialise_system_1(unsigned long int seed)
{
	Chemical_System *cs;
	unsigned int M = 1;  /* number of chemical reactions */
	unsigned int N = 1;  /* number of chemical species */
	
	/* initialise chemical system */
	cs = initialise_chemical_system(M, N, seed);
	
	/* set values for stocastic reaction constants */
 	cs->c[0] = 0.5;
	/* set initial values for species population */
	cs->X[0] = 1000;
	/* set reactant combination function */
	cs->h = h_system_1;
	/* set reaction function */
	cs->R = R_system_1;
	
	return cs;
}

/* reaction 29 */

unsigned int h_system_2(Chemical_System *cs, const unsigned int i)
{
	switch(i) {
		case 1:
			return cs->X[0];
			break;
	
		case 2:
			return cs->X[0] * (cs->X[0] - 1) / 2.0;
			break;
	
		default:
			errx(1, "out-of-range subscript in combination function");
			break;
	}
}

void R_system_2(Chemical_System *cs, const unsigned int i)
{
	switch(i) {
		case 1:
			cs->X[0]++;
			break;
	
		case 2:
			cs->X[0] -= 2;
			break;
	
		default:
			errx(1, "out-of-range subscript %d in reaction function", i);
			break;
	}
}

Chemical_System *initialise_system_2(unsigned long int seed)
{
	Chemical_System *cs;
	unsigned int M = 2;  /* number of chemical reactions */
	unsigned int N = 1;  /* number of chemical species */
	
	/* initialise chemical system */
	cs = initialise_chemical_system(M, N, seed);
	
	/* set values for stocastic reaction constants */
 	cs->c[0] = 5;
 	cs->c[1] = 0.005;
	/* set initial values for species population */
	cs->X[0] = 10;
	/* set reactant combination function */
	cs->h = h_system_2;
	/* set reaction function */
	cs->R = R_system_2;
	
	return cs;
}

/* lotka reaction (reaction 38) */

unsigned int h_system_3(Chemical_System *cs, const unsigned int i)
{
	switch(i) {
		case 1:
			return cs->X[0];
			break;
	
		case 2:
			return cs->X[0] * cs->X[1];
			break;
	
		case 3:
			return cs->X[1];
			break;
	
		default:
			errx(1, "out-of-range subscript in combination function");
			break;
	}
}

void R_system_3(Chemical_System *cs, const unsigned int i)
{
	switch(i) {
		case 1:
			cs->X[0]++;
			break;
	
		case 2:
			cs->X[0]--;
			cs->X[1]++;
			break;
	
		case 3:
			cs->X[1]--;
			break;
	
		default:
			errx(1, "out-of-range subscript %d in reaction function", i);
			break;
	}
}

Chemical_System *initialise_system_3(unsigned long int seed)
{
	Chemical_System *cs;
	unsigned int M = 3;  /* number of chemical reactions */
	unsigned int N = 2;  /* number of chemical species */
	
	/* initialise chemical system */
	cs = initialise_chemical_system(M, N, seed);
	
	/* set values for stocastic reaction constants */
 	cs->c[0] = 10.0;
 	cs->c[1] = 0.01;
 	cs->c[2] = 10.0;
	/* set initial values for species population */
	cs->X[0] = 1000;
	cs->X[1] = 1000;
	/* set reactant combination function */
	cs->h = h_system_3;
	/* set reaction function */
	cs->R = R_system_3;
	
	return cs;
}

/*************/
/* main loop */
/*************/

int main(int argc, char **argv)
{
	Chemical_System *cs;
	FILE *fp;  /* file to save data to */
	double timespan;  /* time to run simulation for */
	unsigned long int seed;  /* random number generator seed */
	
	/* parse command line */
	if(argc < 3 || argc > 4)
		errx(1, "usage: gillespie OUTFILE TIMESPAN [SEED]");
	else {
		if(!(fp = fopen(argv[1], "w")))
			errx(1, "couldn't open output file");
		timespan = atof(argv[2]);
		if(argc == 3)
			seed = time(NULL);
		else
			seed = atol(argv[3]);
	}
		
	/* initialise and create the chemical system */	
	/* uncomment/comment lines below to simulate different chemical systems */
//	cs = initialise_system_1(seed);
//	cs = initialise_system_2(seed);
	cs = initialise_system_3(seed);
	
	/* iterate system */
	output_chemical_system(cs, fp);
	while(cs->t <= timespan) {
		gillespie_algorithm(cs);
		output_chemical_system(cs, fp);
	}

	fclose(fp);
	
	return EXIT_SUCCESS;
}
